from .metrics import *

import unittest


class Binary_detection_tester(unittest.TestCase):
    def test_unsorted(self):
        self.assertRaises(AssertionError, Binary_detection, 10, [2, 3, 4], [3, 4, 2])
        self.assertRaises(AssertionError, Binary_detection, 10, [3, 4, 2], [2, 3, 4])
        self.assertRaises(AssertionError, Binary_detection, 10, [[1, 8]], [[5, 6], [1, 2]])
        self.assertRaises(AssertionError, Binary_detection, 10, [[5, 6], [1, 2]], [[1, 8]])

    def test_nonunique(self):
        self.assertRaises(AssertionError, Binary_detection, 10, [2, 4, 4], [2, 3, 4])
        self.assertRaises(AssertionError, Binary_detection, 10, [2, 3, 4], [2, 4, 4])

    def test_long_anom(self):
        self.assertRaises(AssertionError, Binary_detection, 4, [1], [2, 3, 4])
        self.assertRaises(AssertionError, Binary_detection, 4, [[2, 4]], [1])
        self.assertRaises(AssertionError, Binary_detection, 4, [-1], [1])

    def test_point_to_seq(self):
        anom1 = [3, 4, 5, 7, 8, 11]
        anom2 = [[3, 5], [7, 8], [11, 11]]
        d = Binary_detection(12, anom1, anom2)

        self.assertTrue(np.array_equal(np.array(anom1), d.get_predicted_anomalies_ptwise()))
        self.assertTrue(np.array_equal(np.array(anom2), d.get_gt_anomalies_segmentwise()))

    def test_anomaly_full_seires(self):
        anom1 = [3, 4, 5, 7, 8, 11]
        d = Binary_detection(12, anom1, anom1)

        self.assertTrue(
            np.array_equal(d.get_gt_anomalies_full_series(), np.array([0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 1]))
        )
        self.assertTrue(
            np.array_equal(d.get_predicted_anomalies_full_series(), np.array([0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 0, 1]))
        )

    def test_empty_anom(self):
        anom1 = [3, 4, 5, 7, 8]
        anom2 = []

        d = Binary_detection(12, anom1, anom2)
        self.assertEqual(0, len(d.get_predicted_anomalies_ptwise()))
        self.assertEqual(0, len(d.get_predicted_anomalies_segmentwise()))


class Confusion_metrics_tester(unittest.TestCase):
    def test_metrics(self):
        self.assertEqual(0.75, recall(tp=3, fn=1))
        self.assertEqual(0.75, precision(tp=3, fp=1))
        self.assertEqual(0.6, f1_score(tp=3, fn=1, fp=3))
        self.assertEqual(0.625, f1_from_pr(p=1, r=0.25, beta=0.5))
        self.assertAlmostEqual(5 / 9, f1_from_pr(p=1, r=0.5, beta=2))

    def test_requires_names(self):
        self.assertRaises(TypeError, recall, 3, 4)
        self.assertRaises(TypeError, precision, 3, 4)
        self.assertRaises(TypeError, f1_score, 3, 4, 5)

    def test_zerodivision(self):
        self.assertEqual(0, recall(tp=0, fn=0))
        self.assertEqual(0, precision(tp=0, fp=0))
        self.assertEqual(0, f1_score(tp=0, fp=1, fn=1))


class Metrics_tester(unittest.TestCase):
    def test_PW(self):
        pw = Pointwise_metrics(10, [1, 2, 3, 4], [4, 5, 6])

        self.assertEqual(pw.tp, 1)
        self.assertEqual(pw.fp, 2)
        self.assertEqual(pw.fn, 3)

    def test_PA(self):
        pa = PointAdjust(10, [1, 2, 3, 4, 9], [4, 5, 6])
        self.assertEqual(pa.tp, 4)
        self.assertEqual(pa.fp, 2)
        self.assertEqual(pa.fn, 1)

        pa = PointAdjust(10, [1, 2, 3, 4, 5, 6, 7], [4])
        self.assertEqual(pa.get_score(), 1)

        pa = PointAdjust(10, [1, 2, 3, 4, 5, 6, 7], [9])
        self.assertEqual(pa.get_score(), 0)

    def test_dtPA(self):
        pa = DelayThresholdedPointAdjust(10, [1, 2, 3, 4, 9], [4, 5, 6], k=3)
        self.assertEqual(pa.tp, 4)
        self.assertEqual(pa.fp, 2)
        self.assertEqual(pa.fn, 1)

        pa = DelayThresholdedPointAdjust(10, [1, 2, 3, 4, 9], [4, 5, 6], k=2)
        self.assertEqual(pa.tp, 0)
        self.assertEqual(pa.fp, 2)
        self.assertEqual(pa.fn, 5)

        pa = DelayThresholdedPointAdjust(10, [1, 2, 3, 4, 5, 6, 7], [4], k=3)
        self.assertEqual(pa.get_score(), 1)

        pa = DelayThresholdedPointAdjust(10, [1, 2, 3, 4, 5, 6, 7], [4], k=2)
        self.assertEqual(pa.get_score(), 0)

    def test_pakf(self):
        pa = PointAdjustKPercent(10, [1, 2, 3, 4, 9], [4, 5, 6], k=0.5)
        self.assertEqual(pa.tp, 1)
        self.assertEqual(pa.fp, 2)
        self.assertEqual(pa.fn, 4)

        pa = PointAdjustKPercent(10, [1, 2, 3, 4, 9], [4, 5, 6], k=0.1)
        self.assertEqual(pa.tp, 4)
        self.assertEqual(pa.fp, 2)
        self.assertEqual(pa.fn, 1)

    def test_lspa(self):
        pa = LatencySparsityAware(10, [2, 3, 4, 5, 9], [4, 7], tw=1)
        self.assertAlmostEqual(pa.get_score(), f1_score(tp=pa.tp, fn=pa.fn, fp=pa.fp), 4)
        self.assertEqual(pa.tp, 2)
        self.assertEqual(pa.fp, 1)
        self.assertEqual(pa.fn, 3)

        pa = LatencySparsityAware(10, [2, 3, 4, 5, 9], [4, 7], tw=2)
        self.assertAlmostEqual(pa.get_score(), f1_score(tp=pa.tp, fn=pa.fn, fp=pa.fp), 4)
        self.assertEqual(pa.tp, 1)
        self.assertEqual(pa.fp, 1)
        self.assertEqual(pa.fn, 2)

    def test_Segment(self):
        s = Segmentwise_metrics(10, [[1, 2], [4, 4], [7, 9]], [[0, 6]])
        self.assertEqual(s.tp, 2)
        self.assertEqual(s.fp, 0)
        self.assertEqual(s.fn, 1)

        s = Segmentwise_metrics(10, [[1, 2], [4, 4], [7, 9]], [[6, 6], [8, 8]])
        self.assertEqual(s.tp, 1)
        self.assertEqual(s.fp, 1)
        self.assertEqual(s.fn, 2)

        s = Segmentwise_metrics(10, [[1, 2], [4, 4], [7, 9]], [])
        self.assertEqual(s.tp, 0)
        self.assertEqual(s.fp, 0)
        self.assertEqual(s.fn, 3)

        s = Segmentwise_metrics(10, [[1, 2], [4, 4], [7, 9]], [[0, 9]])
        self.assertEqual(s.tp, 3)
        self.assertEqual(s.fp, 0)
        self.assertEqual(s.fn, 0)

    def test_CF(self):
        c = Composite_f(10, [0, 2, 3, 5, 7, 9], [3, 6])
        f = c.get_score()
        self.assertEqual(c.p, 0.5)
        self.assertEqual(c.r, 0.2)

    def test_affiliation(self):
        a = Affiliation(10, [2, 3], [2])
        f = a.get_score()
        self.assertEqual(a.p, 1)
        self.assertTrue(a.r < 1)

        a = Affiliation(10, [2, 3], [2, 3, 4])
        f = a.get_score()
        self.assertTrue(a.p < 1)
        self.assertEqual(a.r, 1)

    def test_range_pr(self):
        r = Range_PR(10, [2, 3], [2])
        f = r.get_score()
        self.assertEqual(r.p, 1)
        self.assertTrue(r.r < 1)

        r2 = Range_PR(10, [2, 3], [2, 3])
        f2 = r2.get_score()
        self.assertTrue(f2 > f)

        r = Range_PR(10, [2, 3], [2, 3, 4])
        f = r.get_score()
        self.assertTrue(r.p < 1)
        self.assertEqual(r.r, 1)

    def test_NAB(self):
        n = NAB_score(10, [[3, 6]], [3])
        self.assertAlmostEqual(n.get_score(), 100)

        n = NAB_score(10, [[3, 6]], [])
        self.assertAlmostEqual(n.get_score(), 0)

        n = NAB_score(10, [[3, 6]], [1])
        self.assertAlmostEqual(n.get_score(), -100 * 0.11 / 2)

        n = NAB_score(10, [3, 6], [1])
        self.assertTrue(np.isnan(n.get_score()))

    def test_ttol(self):
        t = Time_Tolerant(10, [3, 4, 8], [1, 2, 3], d=2)
        self.assertAlmostEqual(t.recall(), 2 / 3)
        self.assertAlmostEqual(t.precision(), 1)

        t = Time_Tolerant(10, [4, 5], [6], d=1)
        self.assertAlmostEqual(t.recall(), 1 / 2)
        self.assertAlmostEqual(t.precision(), 1)

    def test_TaF(self):
        t = TaF(10, [4, 5, 6], [4, 5, 6])
        self.assertEqual(t.get_score(), 1)

        t = TaF(10, [4, 5, 6], [1, 2, 3])
        self.assertEqual(t.get_score(), 0)

        t = TaF(10, [4, 5, 6], [7, 8, 9])
        self.assertEqual(t.get_score(), 0)
        t = TaF(10, [4, 5, 6], [7, 8, 9], delta=1)
        self.assertTrue(t.get_score() > 0)

        t1 = TaF(10, [4, 5, 8, 9], [4, 5])
        t2 = TaF(10, [4, 5, 8, 9], [5, 8])
        self.assertTrue(t1.get_score() < t2.get_score())

    def test_eTaF(self):
        t = eTaF(10, [4, 5, 6], [4, 5, 6])
        self.assertEqual(t.get_score(), 1)

        t = eTaF(10, [4, 5, 6], [1, 2, 3])
        self.assertEqual(t.get_score(), 0)

        t = eTaF(10, [4, 5, 6], [7, 8, 9])
        self.assertTrue(t.get_score() == 0)

        t1 = eTaF(10, [4, 5, 8, 9], [4, 5])
        t2 = eTaF(10, [4, 5, 8, 9], [5, 8])
        self.assertTrue(t1.get_score() < t2.get_score())

    def test_temp_dist(self):
        t = Temporal_Distance(10, [4, 5, 6], [4, 5, 6])
        self.assertEqual(t.get_score(), 0)

        t = Temporal_Distance(10, [4, 6], [4, 5, 6])
        self.assertEqual(t.get_score(), 1)

        t = Temporal_Distance(10, [4], [4, 5, 6])
        self.assertEqual(t.get_score(), 3)

        t = Temporal_Distance(10, [4, 5, 6], [8])
        self.assertEqual(t.get_score(), 11)

        t = Temporal_Distance(10, [4, 5, 6], [])
        self.assertEqual(t.get_score(), 30)


class Threshold_metric_tester(unittest.TestCase):
    #    def test_roc(self):
    #        a = aucroc(true = [0,0,1,1], score = [0.1,0.4,0.35,0.8])
    def test_auc_pr(self):
        gt = [[2, 3]]
        anomaly_score = [1, 3, 2, 4]
        auc_pr = AUC_PR_pw(gt, anomaly_score)

        score = auc_pr.get_score()
        self.assertAlmostEqual(score, 0.83, 2)

        anomaly_score = [1, 2, 3, 4]
        auc_pr = AUC_PR_pw(gt, anomaly_score)
        score = auc_pr.get_score()
        self.assertEqual(score, 1)

        anomaly_score = [4, 3, 1, 1]
        auc_pr = AUC_PR_pw(gt, anomaly_score)
        score = auc_pr.get_score()
        self.assertEqual(score, 0.5)

    def test_auc_roc(self):
        gt = [[2, 3]]
        anomaly_score = [1, 3, 2, 4]
        auc_roc = AUC_ROC(gt, anomaly_score)

        score = auc_roc.get_score()
        self.assertAlmostEqual(score, 0.75, 2)

        anomaly_score = [1, 2, 3, 4]
        auc_roc = AUC_ROC(gt, anomaly_score)
        score = auc_roc.get_score()
        self.assertEqual(score, 1)

        anomaly_score = [4, 4, 4, 4]
        auc_roc = AUC_ROC(gt, anomaly_score)
        score = auc_roc.get_score()
        self.assertEqual(score, 0.5)

    def test_vus_pr(self):
        gt = [[0, 1]]
        anomaly_score = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        vus_pr = VUS_PR(gt, anomaly_score, max_window=4)

        score = vus_pr.get_score()
        self.assertTrue(score <= 0.2)

        gt = [[1, 3]]
        anomaly_score = [8, 0, 9, 1, 7, 2, 3, 4, 5, 6]
        vus_pr = VUS_PR(gt, anomaly_score, max_window=4)
        score = vus_pr.get_score()
        self.assertTrue(score > 0.5)
        vus_pr = VUS_PR(gt, anomaly_score, max_window=0)
        score = vus_pr.get_score()
        self.assertTrue(score < 0.5)

    def test_vus_roc(self):
        gt = [[0, 1]]
        anomaly_score = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        vus = VUS_ROC(gt, anomaly_score, max_window=4)

        score = vus.get_score()
        self.assertTrue(score <= 0.1)

        gt = [[1, 3]]
        anomaly_score = [8, 0, 9, 1, 7, 2, 3, 4, 5, 6]
        vus = VUS_ROC(gt, anomaly_score, max_window=4)
        score = vus.get_score()
        self.assertTrue(score > 0.4)
        vus = VUS_ROC(gt, anomaly_score, max_window=0)
        score = vus.get_score()
        self.assertTrue(score < 0.4)

    def test_PatK(self):
        gt = [[2, 3]]

        anomaly_score = [1, 4, 2, 3]
        patk = PatK_pw(gt, anomaly_score)
        score = patk.get_score()
        self.assertEqual(score, 0.5)

        anomaly_score = [1, 2, 3, 4]
        patk = PatK_pw(gt, anomaly_score)
        score = patk.get_score()
        self.assertEqual(score, 1)

        anomaly_score = [3, 4, 1, 2]
        patk = PatK_pw(gt, anomaly_score)
        score = patk.get_score()
        self.assertEqual(score, 0)

        anomaly_score = [3, 4, 1, 2]
        patk = PatK_pw([1, 2, 3], anomaly_score)
        score = patk.get_score()
        self.assertAlmostEqual(score, 2 / 3)

        anomaly_score = [2, 1, 1, 0]
        patk = PatK_pw([0, 1], anomaly_score)
        score = patk.get_score()
        self.assertAlmostEqual(score, 2 / 3)

        patk = PatK_pw([], [0, 1, 2, 4])
        self.assertRaises(AssertionError, patk.get_score)

    def test_best_threshold_pw(self):
        gt = [[2, 3]]

        anomaly_score = [1, 3, 2, 4]
        metric = Best_threshold_pw(gt, anomaly_score)
        score = metric.get_score()
        self.assertAlmostEqual(score, 2 * 2 / 3 * 1 / (1 + 2 / 3))

        anomaly_score = [2, 3, 1, 4]
        metric = Best_threshold_pw(gt, anomaly_score)
        score = metric.get_score()
        self.assertAlmostEqual(score, 2 * 1 / 2 * 1 / (1 + 1 / 2))

        anomaly_score = [4, 3, 1, 2]
        metric = Best_threshold_pw(gt, anomaly_score)
        score = metric.get_score()
        self.assertAlmostEqual(score, 2 * 1 / 2 * 1 / (1 + 1 / 2))


if __name__ == "__main__":
    unittest.main()



# from maketable import *


# class Score_graphs_table(Table):
#     def __init__(self, metric_names, results, marks=[]):
#         self.metric_names = metric_names
#         self.results = results
#         self.marks = marks
#         super().__init__(Table_content([], [], []), scale=2)
#         self.x_factor = 1 / 20 * self.scale
#         self.y_factor = 1 / 2
#         self.y_shift = -0.2

#         self.row_length = 2
#         self.n_rows = len(metric_names)

#     def add_top_row(self):
#         self.string += "Metric"
#         self.string += "&"
#         self.string += "Score"
#         self.end_row()

#     def add_next_row(self):
#         self.string += self.metric_names[self.rows_added]
#         self.string += "&"
#         self.add_graph(self.rows_added + 1)
#         self.end_row()
#         self.rows_added += 1

#     def add_graph(self, number):
#         self.add_line(f"\\begin{{tikzpicture}}[baseline=-\\the\\dimexpr\\fontdimen22\\textfont2\\relax]")
#         for x in self.marks:
#             self.add_line(
#                 f"\draw[-, gray] ({x*self.x_factor},{self.y_shift}) -- ({x*self.x_factor},{0.2*self.y_factor + self.y_shift});"
#             )
#         self.add_line(
#             f"\draw[-, gray] (0,{self.y_shift}) -- ({self.x_factor*(len(self.results[self.metric_names[number-1]])-1)},{self.y_shift});"
#         )
#         self.add_line("\\foreach \\i/\\a in")
#         self.add_line(
#             str(
#                 [
#                     (round(i * self.x_factor, 3), round(a * self.y_factor + self.y_shift, 3))
#                     for i, a in enumerate(self.results[self.metric_names[number - 1]])
#                 ]
#             )
#             .replace(",", "/")
#             .replace(")/", ",")
#             .replace("(", "")
#             .replace("[", "{")
#             .replace("]", "}")
#             .replace(")}", "}{")
#         )
#         self.add_line("\\coordinate (now) at (\\i,\\a) {};")
#         self.add_line("  \\ifthenelse{\\equal{\\i}{0.0}}{}{")
#         self.add_line("  \\draw[-, teal, thick] (prev) -- (now);")
#         self.add_line("  }")
#         self.add_line("  \\coordinate (prev) at (\\i,\\a) {};")
#         self.add_line("}")
#         self.add_line("\\end{tikzpicture}")


# def score_graphs():
#     # define scenario
#     ts_length = 100
#     pred_length = 5
#     gt_length = 20
#     gt_start = 40
#     marks = [35, 40, 55, 60]
#     assert pred_length % 2 == 1

#     # prepare metrics list
#     All_metrics.remove(metrics.Range_PR)
#     all_metrics_and_rffront = [*All_metrics, metrics.Range_PR, Range_PR_front]

#     # make results and names
#     metric_names = []
#     result = {}
#     for metric in all_metrics_and_rffront:
#         # set names
#         if metric == metrics.TaF:
#             metric_names.append(metric(5, [3, 4], [3], delta=10).name)
#         elif metric == metrics.Time_Tolerant:
#             metric_names.append(metric(5, [3, 4], [3], d=10).name)
#         else:
#             metric_names.append(metric(5, [3, 4], [3]).name)

#         # get results
#         current_result = []
#         for pred_mid in range(pred_length // 2, ts_length - pred_length // 2):
#             gt = [[gt_start, gt_start + gt_length - 1]]
#             pred = [[pred_mid - pred_length // 2, pred_mid + pred_length // 2]]
#             if metric == metrics.TaF:
#                 current_result.append(metric(ts_length, gt, pred, delta=10).get_score())
#             elif metric == metrics.Time_Tolerant:
#                 current_result.append(metric(ts_length, gt, pred, d=10).get_score())
#             else:
#                 current_result.append(metric(ts_length, gt, pred).get_score())
#         current_result = np.array(current_result)
#         current_result = (current_result - min(current_result)) / (max(current_result) - min(current_result))
#         result[metric_names[-1]] = current_result

#     # make the table
#     table = Score_graphs_table(metric_names, result, marks)
#     table.write()
#     print(table)


# score_graphs()



from typing import Any

import numpy as np

from prts.base.time_series_metrics import BaseTimeSeriesMetrics


class TimeSeriesPrecisionRecall(BaseTimeSeriesMetrics):
    """ This class calculates precision and recall for time series """

    def score(self, real: np.ndarray, pred: np.ndarray) -> Any:
        """

        Args:
            real:
            pred:

        Returns:

        """
        # TODO: impl



import argparse
from typing import Callable
import math
import copy
from .DataManage import File_IO, Range as rng#, Time_Plot


class TaPR:
    def __init__(self, theta: float, delta: int, past_range: bool = False):
        self._past_range = past_range
        self._predictions = []  # list of Ranges
        self._anomalies = []    # list of Ranges
        self._ambiguous_inst = [] # list of Ranges

        self._set_predictions = False
        self._set_anomalies = False

        #self._rho = theta
        #self._pi = theta
        self._theta = theta
        self._delta = delta

        pass

    def set_anomalies(self, anomaly_list: list) -> None:
        self._anomalies = copy.deepcopy(anomaly_list)
        self._gen_ambiguous()
        self._set_anomalies = True

    def set_predictions(self, prediction_list: list) -> None:
        self._predictions = copy.deepcopy(prediction_list)
        self._set_predictions = True

    def _gen_ambiguous(self):
        for i in range(len(self._anomalies)):
            start_id = self._anomalies[i].get_time()[1] + 1
            end_id = end_id = start_id + self._delta

            if self._past_range:
                end_id = start_id + int(self._delta * (self._anomalies[i].get_time()[1] - self._anomalies[i].get_time()[0]))
            
            #if the next anomaly occurs during the theta, update the end_id
            if i+1 < len(self._anomalies) and end_id > self._anomalies[i+1].get_time()[0]:
                end_id = self._anomalies[i+1].get_time()[0] - 1

            if start_id > end_id:
                start_id = -2
                end_id = -1

            self._ambiguous_inst.append(rng.Range(start_id, end_id, str(i)))

    def get_n_predictions(self):
        return len(self._predictions)

    def get_n_anomalies(self):
        return len(self._anomalies)

    def _ids_2_objects(self, id_list, range_list):
        result = []
        for id in id_list:
            result.append(range_list[id])
        return result

    def TaR_d(self) -> float and list:
        score, detected_id_list = self._TaR_d(self._anomalies, self._ambiguous_inst, self._predictions, self._theta)
        return score, self._ids_2_objects(detected_id_list, self._anomalies)

    def _TaR_d(self, anomalies: list, ambiguous_inst: list, predictions: list, threshold: float) -> float and list:
        total_score = 0.0
        detected_anomalies = []
        total_score_p = 0.0
        for anomaly_id in range(len(anomalies)):
            anomaly = anomalies[anomaly_id]
            ambiguous = ambiguous_inst[anomaly_id]

            max_score = self._sum_of_func(anomaly.get_time()[0], anomaly.get_time()[1],
                                          anomaly.get_time()[0], anomaly.get_time()[1], self._uniform_func)

            score = 0.0
            for prediction in predictions:
                score += self._overlap_and_subsequent_score(anomaly, ambiguous, prediction)
            
            total_score_p += min(1.0, score/max_score)
            if min(1.0, score / max_score) >= threshold:
                total_score += 1.0
                detected_anomalies.append(anomaly_id)

        if len(anomalies) == 0:
            self.TaR_p_value = 0
            return 0.0, []
        else:
            self.TaR_p_value = total_score_p / len(anomalies)
            return total_score / len(anomalies), detected_anomalies
    
    def TaR_p(self) -> float:
        total_score = 0.0
        for anomaly_id in range(len(self._anomalies)):
            anomaly = self._anomalies[anomaly_id]
            ambiguous = self._ambiguous_inst[anomaly_id]

            max_score = self._sum_of_func(anomaly.get_time()[0], anomaly.get_time()[1],
                                          anomaly.get_time()[0], anomaly.get_time()[1], self._uniform_func)

            score = 0.0
            for prediction in self._predictions:
                score += self._overlap_and_subsequent_score(anomaly, ambiguous, prediction)

            total_score += min(1.0, score/max_score)

        if len(self._anomalies) == 0:
            return 0.0
        else:
            return total_score / len(self._anomalies)


    def TaP_d(self) -> float and list:
        score, correct_id_list = self._TaP_d(self._anomalies, self._ambiguous_inst, self._predictions, self._theta)
        return score, self._ids_2_objects(correct_id_list, self._predictions)

    def _TaP_d(self, anomalies, ambiguous_inst, predictions, threshold):
        #Compute TaP_d and TaP_p in one function to optimize the performance
        correct_predictions = []
        total_score = 0.0
        total_score_p = 0.0
        for prediction_id in range(len(predictions)):
            max_score = predictions[prediction_id].get_time()[1] - predictions[prediction_id].get_time()[0] + 1

            score = 0.0
            for anomaly_id in range(len(anomalies)):
                anomaly = anomalies[anomaly_id]
                ambiguous = ambiguous_inst[anomaly_id]

                score += self._overlap_and_subsequent_score(anomaly, ambiguous, predictions[prediction_id])
            total_score_p += score / max_score
            if (score/max_score) >= threshold:
                total_score += 1.0
                correct_predictions.append(prediction_id)

        if len(predictions) == 0:
            self.TaP_p_value = 0
            return 0.0, []
            
        else:
            self.TaP_p_value = total_score_p / len(predictions)
            return total_score / len(predictions), correct_predictions
        
    
        
    def _detect(self, src_range: rng.Range, ranges: list, theta: int) -> bool:
        rest_len = src_range.get_time()[1] - src_range.get_time()[0] + 1
        for dst_range in ranges:
            len = self._overlapped_len(src_range, dst_range)
            if len != -1:
                rest_len -= len
        return (float)(rest_len) / (src_range.get_time()[1] - src_range.get_time()[0] + 1) <= (1.0 - theta)

    def _overlapped_len(self, range1: rng.Range, range2: rng.Range) -> int:
        detected_start = max(range1.get_time()[0], range2.get_time()[0])
        detected_end = min(range1.get_time()[1], range2.get_time()[1])

        if detected_end < detected_start:
            return 0
        else:
            return detected_end - detected_start + 1

    def _min_max_norm(self, value: int, org_min: int, org_max: int, new_min: int, new_max: int) -> float:
        if org_min == org_max:
            return new_min
        else:
            return (float)(new_min) + (float)(value - org_min) * (new_max - new_min) / (org_max - org_min)

    def _decaying_func(self, val: float) -> float:
        assert (-6 <= val <= 6)
        return 1 / (1 + math.exp(val))

    def _ascending_func(self, val: float) -> float:
        assert (-6 <= val <= 6)
        return 1 / (1 + math.exp(val * -1))

    def _uniform_func(self, val: float) -> float:
        return 1.0

    def _sum_of_func(self, start_time: int, end_time: int, org_start: int, org_end: int,
                     func: Callable[[float], float]) -> float:
        val = 0.0
        for timestamp in range(start_time, end_time + 1):
            val += func(self._min_max_norm(timestamp, org_start, org_end, -6, 6))
        return val

    def _overlap_and_subsequent_score(self, anomaly: rng.Range, ambiguous: rng.Range, prediction: rng.Range) -> float:
        score = 0.0

        detected_start = max(anomaly.get_time()[0], prediction.get_time()[0])
        detected_end = min(anomaly.get_time()[1], prediction.get_time()[1])

        score += self._sum_of_func(detected_start, detected_end,
                                   anomaly.get_time()[0], anomaly.get_time()[1], self._uniform_func)

        if ambiguous.get_time()[0] < ambiguous.get_time()[1]:
            detected_start = max(ambiguous.get_time()[0], prediction.get_time()[0])
            detected_end = min(ambiguous.get_time()[1], prediction.get_time()[1])

            score += self._sum_of_func(detected_start, detected_end,
                                       ambiguous.get_time()[0], ambiguous.get_time()[1], self._decaying_func)

        return score

    def TaR_p(self) -> float:
        total_score = 0.0
        for anomaly_id in range(len(self._anomalies)):
            anomaly = self._anomalies[anomaly_id]
            ambiguous = self._ambiguous_inst[anomaly_id]

            max_score = self._sum_of_func(anomaly.get_time()[0], anomaly.get_time()[1],
                                          anomaly.get_time()[0], anomaly.get_time()[1], self._uniform_func)

            score = 0.0
            for prediction in self._predictions:
                score += self._overlap_and_subsequent_score(anomaly, ambiguous, prediction)

            total_score += min(1.0, score/max_score)

        if len(self._anomalies) == 0:
            return 0.0
        else:
            return total_score / len(self._anomalies)

    def TaP_p(self) -> float:
        total_score = 0.0
        for prediction in self._predictions:
            max_score = prediction.get_time()[1] - prediction.get_time()[0] + 1

            score = 0.0
            for anomaly_id in range(len(self._anomalies)):
                anomaly = self._anomalies[anomaly_id]
                ambiguous = self._ambiguous_inst[anomaly_id]

                score += self._overlap_and_subsequent_score(anomaly, ambiguous, prediction)

            total_score += score/max_score

        if len(self._predictions) == 0:
            return 0.0
        else:
            return total_score / len(self._predictions)


def compute(anomalies: list, predictions: list, alpha: float, theta: float, delta: int) -> dict:
    ev = TaPR(theta, delta)

    ev.set_anomalies(anomalies)
    ev.set_predictions(predictions)

    tard_value, detected_list = ev.TaR_d()
    tarp_value = ev.TaR_p()

    tapd_value, correct_list = ev.TaP_d()
    tapp_value = ev.TaP_p()

    result = {}
    tar_value = alpha * tard_value + (1 - alpha) * tarp_value
    result['TaR'] = tar_value
    result['TaRd'] = tard_value
    result['TaRp'] = tarp_value

    tap_value = alpha * tapd_value + (1 - alpha) * tapp_value
    result['TaP'] = tap_value
    result['TaPd'] = tapd_value
    result['TaPp'] = tapp_value

    detected_anomalies = []
    for value in detected_list:
        detected_anomalies.append(value.get_name())

    result['Detected_Anomalies'] = detected_anomalies
    result['Detected_Anomalies_Ranges'] = detected_list
    result['Correct_Predictions_Ranges'] = correct_list

    if tar_value + tap_value == 0:
        result['f1'] = 0.0
    else:
        result['f1'] = (2 * tar_value * tap_value) / (tar_value + tap_value)

    return result


def compute_with_load(anomaly_file: str, prediction_file: str, file_type: str, alpha: float, theta: float, delta: int) -> dict:
    anomalies = File_IO.load_file(anomaly_file, file_type)
    predictions = File_IO.load_file(prediction_file, file_type)
    return compute(anomalies, predictions, alpha, theta, delta)


def print_result(anomalies: list, predictions: list, alpha: float, theta: float, delta: int, verbose: bool, graph: str) -> None:
    org_predictions = copy.deepcopy(predictions)
    result = compute(anomalies, predictions, alpha, theta, delta)

    print("The parameters (alpha, theta, delta) are set as %g, %g, and %d." % (alpha, theta, delta))

    print('\n[TaR]:', "%0.5f" % result['TaR'])
    print("\t* Detection score:", "%0.5f" % result['TaRd'])
    print("\t* Portion score:", "%0.5f" % result['TaRp'])
    if verbose:
        buf = '\t\tdetected anomalies: '
        if len(result['Detected_Anomalies_Ranges']) == 0:
            buf += "None  "
        else:
            for value in result['Detected_Anomalies_Ranges']:
                buf += value.get_name() + '(' + str(value.get_time()[0]) + ':' + str(value.get_time()[1]) + '), '
        print(buf[:-2])


    print('\n[TaP]:', "%0.5f" % result['TaP'])
    print("\t* Detection score:", "%0.5f" % result['TaPd'])
    print("\t* Portion score:", "%0.5f" % result['TaPp'])
    if verbose:
        buf = '\t\tcorrect predictions: '
        if len(result['Correct_Predictions_Ranges']) == 0:
            buf += "None  "
        else:
            for value in result['Correct_Predictions_Ranges']:
                buf += value.get_name() + '(' + str(value.get_time()[0]) + ':' + str(value.get_time()[1]) + '), '
        print(buf[:-2])


    assert(graph == 'screen' or graph == 'file' or graph == 'none' or graph == 'all')
    if graph == 'screen' or graph == 'file' or graph == 'all':
        Time_Plot.draw_graphs(anomalies, org_predictions, graph)


if __name__ == '__main__':
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("--anomalies", help="anomaly file name (ground truth)", required=True)
    argument_parser.add_argument("--predictions", help="prediction file name", required=True)
    argument_parser.add_argument("--filetype", help="choose the file type between range and stream", required=True)
    argument_parser.add_argument("--graph", help="show graph of results")

    argument_parser.add_argument("--verbose", help="show detail results", action='store_true')
    argument_parser.add_argument("--theta", help="set parameter theta")
    argument_parser.add_argument("--alpha", help="set parameter alpha")
    argument_parser.add_argument("--delta", help="set parameter delta")
    arguments = argument_parser.parse_args()

    arguments = argument_parser.parse_args()
    theta, alpha, delta, graph = 0.5, 0.8, 600, 'none'  #default values
    if arguments.theta is not None:
        theta = float(arguments.theta)
    if arguments.alpha is not None:
        alpha = float(arguments.alpha)
    if arguments.delta is not None:
        delta = int(arguments.delta)
    if arguments.graph is not None:
        graph = arguments.graph

    assert(0.0 <= theta <= 1.0)
    assert(0.0 <= alpha <= 1.0)
    assert(isinstance(delta, int))
    assert(graph == 'screen' or graph == 'file' or graph == 'none' or graph == 'all')

    anomalies = File_IO.load_file(arguments.anomalies, arguments.filetype)
    predictions = File_IO.load_file(arguments.predictions, arguments.filetype)

    print_result(anomalies, predictions, alpha, theta, delta, arguments.verbose, graph)



from .tapr import *
import math
import argparse
from .DataManage import File_IO#, Time_Plot
import numpy as np
from .DataManage import Range as rng

class eTaPR(TaPR):
    def __init__(self, theta_p, theta_r, delta=0.0):
        super(eTaPR, self).__init__(0.0, 0)
        self._predictions_weight = []
        self._predictions_total_weight = 0.0
        self._prune_predictions = []

        self._theta_p = theta_p
        self._theta_r = theta_r
        self._delta_ratio = delta

        self._overlap_score_mat_org = np.zeros(1)
        self._overlap_score_mat_elm = np.zeros(1) #eleminate by prunning
        self._max_anomaly_score = []
        self._max_prediction_score = []

        self._weight_func = math.sqrt


    def _gen_ambiguous(self):
        for i in range(len(self._anomalies)):
            start_id = self._anomalies[i].get_time()[1] + 1
            end_id = start_id + int(self._delta_ratio * (self._anomalies[i].get_time()[1] - self._anomalies[i].get_time()[0]))

            # if the next anomaly occurs during the theta, update the end_id
            if i + 1 < len(self._anomalies) and end_id > self._anomalies[i + 1].get_time()[0]:
                end_id = self._anomalies[i + 1].get_time()[0] - 1

            if start_id > end_id:
                start_id = -2
                end_id = -1

            self._ambiguous_inst.append(rng.Range(start_id, end_id, str(i)))

    #load data -> build the score matrix -> do pruning
    def set(self, anomalies: list, predictions: list) -> None:
        #loading data
        self.set_anomalies(anomalies)
        self.set_predictions(predictions)

        #computing weights
        for a_prediction in self._predictions:
            first, last = a_prediction.get_time()
            temp_weight = math.sqrt(last-first+1)
            self._predictions_weight.append(temp_weight)
            self._predictions_total_weight += temp_weight

        #computing the score matrix
        self._overlap_score_mat_org = np.zeros((self.get_n_anomalies(), self.get_n_predictions()))
        for anomaly_id in range(self.get_n_anomalies()):
            for prediction_id in range(self.get_n_predictions()):
                self._overlap_score_mat_org[anomaly_id, prediction_id] = \
                    float(self._overlap_and_subsequent_score(self._anomalies[anomaly_id], self._ambiguous_inst[anomaly_id], self._predictions[prediction_id]))

        #computing the maximum scores for each anomaly or prediction
        for an_anomaly in self._anomalies:
            start, end = an_anomaly.get_time()
            self._max_anomaly_score.append(float(self._sum_of_func(start, end, start, end, self._uniform_func)))
        for a_prediction in self._predictions:
            self._max_prediction_score.append(a_prediction.get_len())

        #pruning
        self._pruning()

    def _pruning(self):
        self._overlap_score_mat_elm = self._overlap_score_mat_org.copy()

        while True:
            tars = self._overlap_score_mat_elm.sum(axis=1)/self._max_anomaly_score
            elem_anomaly_ids = set(np.where(tars<self._theta_r)[0]) - set(np.where(tars==0.0)[0])
            for id in elem_anomaly_ids:
                self._overlap_score_mat_elm[id] = np.zeros(self.get_n_predictions())
            taps = self._overlap_score_mat_elm.sum(axis=0)/self._max_prediction_score
            elem_prediction_ids = set(np.where(taps<self._theta_p)[0]) - set(np.where(taps==0.0)[0])
            for id in elem_prediction_ids:
                self._overlap_score_mat_elm[:, id] = np.zeros(self.get_n_anomalies())

            if len(elem_anomaly_ids) == 0 and len(elem_prediction_ids) == 0:
                break

    def _etar_d(self, theta: float) -> np.array and list:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return np.zeros(self.get_n_anomalies()), []

        scores = self._overlap_score_mat_elm.sum(axis=1)/self._max_anomaly_score
        scores = np.where(scores >= theta, 1.0, scores)
        scores = np.where(scores <  theta, 0.0, scores)
        detected_id_list = np.where(scores >= theta)[0]

        return scores, detected_id_list

    def eTaR_d(self) -> float and list:
        _, detected_id_list = self._etar_d(self._theta_r)
        return len(detected_id_list)/self.get_n_anomalies(), detected_id_list

    def _etar_p(self) -> np.array:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        scores = self._overlap_score_mat_elm.sum(axis=1) / self._max_anomaly_score
        scores = np.where(scores > 1.0, 1.0, scores)
        return scores

    def eTaR_p(self) -> float:
        scores = self._etar_p()
        return scores.mean()

    def eTaR(self) -> float:

        detection_scores, detected_id_list = self._etar_d(self._theta_r)
        portion_scores = self._etar_p()

        return ((detection_scores + detection_scores * portion_scores)/2).mean(), portion_scores.mean(), len(detected_id_list)/self.get_n_anomalies(), detected_id_list

    def _etap_d(self, theta: float) -> np.array and list:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0, []

        scores = self._overlap_score_mat_elm.sum(axis=0) / self._max_prediction_score
        scores = np.where(scores >= theta, 1.0, scores)
        scores = np.where(scores <  theta, 0.0, scores)
        correct_id_list = np.where(scores >= theta)[0]

        return scores, correct_id_list

    def eTaP_d(self) -> float and list:
        _, correct_id_list = self._etap_d(self._theta_p)

        tapd = 0.0
        for correct_id in correct_id_list:
            tapd += self._predictions_weight[correct_id]
        tapd /= float(self._predictions_total_weight)

        return tapd, correct_id_list

    def _etap_p(self) -> np.array:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        scores = self._overlap_score_mat_elm.sum(axis=0) / self._max_prediction_score
        return scores

    def eTaP_p(self) -> float:
        scores = self._etap_p()

        final_score = 0.0
        for i in range(len(scores)):
            final_score += float(self._predictions_weight[i]) * scores[i]
        final_score /= float(self._predictions_total_weight)
        return final_score

    def eTaP(self) -> float:
        #Computing etap_d and etap_p manually to optimize the performance
        etap_d = 0
        etap_p = 0
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            etap_d,etap_p = 0.0, 0.0
        
        etap_d = self._overlap_score_mat_elm.sum(axis=0) / self._max_prediction_score
        etap_p = etap_d
        etap_d = np.where(etap_d >= self._theta_p, 1.0, etap_d)
        etap_d = np.where(etap_d <  self._theta_p, 0.0, etap_d)
        corrected_id_list = np.where(etap_d >= self._theta_p)[0]

        detection_scores = etap_d
        portion_scores = etap_p
        eTaP_d,eTaP_p = 0.0,0.0


        scores = (detection_scores + detection_scores * portion_scores)/2
        final_score = 0.0
        for i in range(max(len(scores),len(etap_d),len(corrected_id_list))):
            if i < len(scores):
                final_score += float(self._predictions_weight[i]) * scores[i]
            if i < len(etap_p):
                eTaP_p += float(self._predictions_weight[i]) * etap_p[i]
            if i < len(corrected_id_list):
                eTaP_d += self._predictions_weight[corrected_id_list[i]]
        
        final_score /= float(self._predictions_total_weight)
        eTaP_d /= float(self._predictions_total_weight)
        eTaP_p /= float(self._predictions_total_weight)
        self.eTaP_d_value = eTaP_d
        self.eTaP_p_value = eTaP_p
        self.corrected_id_list = corrected_id_list
        return final_score

    # conventional precision
    def precision(self) -> float:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        return self._overlap_score_mat_org.sum() / sum(self._max_prediction_score)

    # conventional recall
    def recall(self) -> float:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        return self._overlap_score_mat_org.sum() / sum(self._max_anomaly_score)

    # point adjust precision
    def point_adjust_precision(self, theta: float) -> float:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        _, detected_id_list = self._TaR_d(self._anomalies, [ rng.Range(-2, -1, '' ) for i in range(len(self._anomalies)) ], self._predictions, theta)
        
        hit_cnt = 0
        for detected_id in detected_id_list:
            hit_cnt += self._anomalies[detected_id].get_len()

        extended_predictions_len = sum(self._max_prediction_score) + hit_cnt - self._overlap_score_mat_org.sum()
        
        return hit_cnt / extended_predictions_len

    def point_adjust_recall(self, theta: float) -> float:
        if self.get_n_anomalies() == 0.0 or self.get_n_predictions() == 0.0:
            return 0.0

        _, detected_id_list = self._TaR_d(self._anomalies, [ rng.Range(-2, -1, '' ) for i in range(len(self._anomalies)) ], self._predictions, theta)
        hit_cnt = 0
        for detected_id in detected_id_list:
            hit_cnt += self._anomalies[detected_id].get_len()
        return hit_cnt / sum(self._max_anomaly_score)

import time
def evaluate_w_ranges(anomalies: list, predictions: list, theta_p: float, theta_r: float, delta: float = 0.0) -> dict:
    assert(0.0 <= theta_p <= 1.0)
    assert(0.0 <= theta_r <= 1.0)
    assert(0.0 <= delta <= 1.0)
    
    ev = eTaPR(theta_p, theta_r, delta)
    ev.set(anomalies, predictions)
    
    
    tar_value, tarp_value, tard_value, detected_id_list = ev.eTaR()

    tap_value = ev.eTaP()
    tapd_value = ev.eTaP_d_value
    tapp_value = ev.eTaP_p_value
    
    
    result = {}
    result['eTaR'] = tar_value
    result['eTaRd'] = tard_value
    result['eTaRp'] = tarp_value

    result['eTaP'] = tap_value
    result['eTaPd'] = tapd_value
    result['eTaPp'] = tapp_value
    
    # detected_anomalies = []
    # for id in detected_id_list:
    #     detected_anomalies.append(anomalies[id])

    # correct_predictions = []
    # for id in correct_id_list:
    #     correct_predictions.append(predictions[id])
    
    # result['Detected_Anomalies'] = detected_anomalies
    # result['Correct_Predictions'] = correct_predictions
    
    if tar_value + tap_value == 0:
        result['f1'] = 0.0
    else:
        result['f1'] = (2 * tar_value * tap_value) / (tar_value + tap_value)

    # false_alarm = 0
    # false_alarm_cnt = 0
    # for id in range(len(predictions)):
    #     if id not in correct_id_list:
    #         false_alarm += predictions[id].get_len()
    #         false_alarm_cnt += 1
    
    
    # result['False Alarm'] = false_alarm
    # result['N False Alarm'] = false_alarm_cnt
    
    # result['precision'] = ev.precision()
    # result['recall'] = ev.recall()
    # result['point_adjust_precision'] = ev.point_adjust_precision(1e-10)
    # result['point_adjust_recall'] = ev.point_adjust_recall(1e-10)
    return result


def evaluate_w_streams(anomalies: list, predictions: list, theta_p = 0.7, theta_r: float = 0.1, delta: float = 0.0) -> dict:
    assert(0.0 <= theta_p <= 1.0)
    assert(0.0 <= theta_r <= 1.0)
    assert(0.0 <= delta <= 1.0)

    anomalous_ranges = File_IO.load_stream_2_range(anomalies, 0, 1, True)
    predicted_ranges = File_IO.load_stream_2_range(predictions, 0, 1, True)

    return evaluate_w_ranges(anomalies =anomalous_ranges,
                   predictions =predicted_ranges,
                   theta_p=theta_p,
                   theta_r=theta_r,
                   delta=delta)


def evaluate_w_files(anomaly_file: str, prediction_file: str, file_type: str, theta_p: float, theta_r: float, delta: float = 0.0) -> dict:
    assert(0.0 <= theta_p <= 1.0)
    assert(0.0 <= theta_r <= 1.0)
    assert(0.0 <= delta <= 1.0)

    anomalies = File_IO.load_file(anomaly_file, file_type)
    predictions = File_IO.load_file(prediction_file, file_type)

    return evaluate_w_ranges(anomalies, predictions, theta_p, theta_r, delta)


def print_results(result: dict, verbose: bool) -> None:
    print('\n[TaR]:', "%0.5f" % result['TaR'])
    print("\t* Detection score:", "%0.5f" % result['TaRd'])
    print("\t* Portion score:", "%0.5f" % result['TaRp'])
    if verbose:
        buf = '\t\tdetected anomalies: '
        if len(result['Detected_Anomalies']) == 0:
            buf += "None  "
        else:
            for value in result['Detected_Anomalies']:
                buf += value.get_name() + '(' + str(value.get_time()[0]) + ':' + str(value.get_time()[1]) + '), '
        print(buf[:-2])


    print('\n[TaP]:', "%0.5f" % result['TaP'])
    print("\t* Detection score:", "%0.5f" % result['TaPd'])
    print("\t* Portion score:", "%0.5f" % result['TaPp'])
    if verbose:
        buf = '\t\tcorrect predictions: '
        if len(result['Correct_Predictions']) == 0:
            buf += "None  "
        else:
            for value in result['Correct_Predictions']:
                buf += value.get_name() + '(' + str(value.get_time()[0]) + ':' + str(value.get_time()[1]) + '), '
        print(buf[:-2])


def draw_graph(anomalies: list, predictions: list, graph_dst: str) -> None:
    assert (graph_dst == 'screen' or graph_dst == 'file' or graph_dst == 'none' or graph_dst == 'all')
    if graph_dst == 'screen' or graph_dst == 'file' or graph_dst == 'all':
        Time_Plot.draw_graphs(anomalies, predictions, graph_dst)


if __name__ == '__main__':
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("--anomalies", help="anomaly file name (ground truth)", required=True)
    argument_parser.add_argument("--predictions", help="prediction file name", required=True)
    argument_parser.add_argument("--filetype", help="choose the file type between range and stream", required=True)
    argument_parser.add_argument("--graph", help="show graph of results")

    argument_parser.add_argument("--verbose", help="show detail results", action='store_true')
    argument_parser.add_argument("--theta_r", help="set parameter theta_r")
    argument_parser.add_argument("--theta_p", help="set parameter theta_p")
    argument_parser.add_argument("--delta", help="set parameter delta")
    # arguments = argument_parser.parse_args()

    arguments = argument_parser.parse_args()
    theta_p, theta_r, delta, graph = 0.5, 0.1, 0.0, 'none'  #default values
    if arguments.tp is not None:
        theta_p = float(arguments.tp)
    if arguments.tr is not None:
        theta_r = float(arguments.tr)
    if arguments.delta is not None:
        delta = int(arguments.delta)
    if arguments.graph is not None:
        graph = arguments.graph

    # assert(isinstance(delta, int))
    assert(graph == 'screen' or graph == 'file' or graph == 'none' or graph == 'all')

    anomalies = File_IO.load_file(arguments.anomalies, arguments.filetype)
    predictions = File_IO.load_file(arguments.predictions, arguments.filetype)
    results = evaluate_w_ranges(anomalies, predictions, theta_p, theta_r, delta)

    print_results(results, arguments.verbose)
    draw_graph(anomalies, predictions, graph)




from typing import Any

import numpy as np


class BaseTimeSeriesMetrics:
    """Base class for time series metrics """

    def score(self, real: np.ndarray, pred: np.ndarray) -> Any:
        """

        Args:
            real:
            pred:

        Returns:

        """
        ...

    def _udf_gamma(self):
        """The function of the user-defined gamma.

        Returns:
            float: the value of the user-defined gamma
        """

        return 1.0

    def _gamma_select(self, gamma: str, overlap: int) -> float:
        """The function of selecting the gamma value according to the parameters.

        Args:
            gamma: str
                - 'one': the value 1
                - 'reciprocal';: a reciprocal of the overlap
                - 'udf_gamma': user defined gamma
            overlap: int
                overlap between real and pred

        Returns:
            float: the selected gamma value
        """
        assert type(overlap) == int, TypeError("")

        if gamma == "one":
            return 1.0
        elif gamma == "reciprocal":
            if overlap > 1:
                return 1.0 / overlap
            else:
                return 1.0
        elif gamma == "udf_gamma":
            if overlap > 1:
                return 1.0 / self._udf_gamma()
            else:
                return 1.0
        else:
            raise ValueError(f"Expected one of one, reciprocal, udf_gamma. gamma type string: {gamma}")

    def _gamma_function(self, overlap_count):
        overlap = overlap_count[0]
        return self._gamma_select(self.cardinality, overlap)

    def _compute_omega_reward(self, r1, r2, overlap_count):
        if r1[1] < r2[0] or r1[0] > r2[1]:
            return 0
        else:
            overlap_count[0] += 1
            overlap = np.zeros(r1.shape)
            overlap[0] = max(r1[0], r2[0])
            overlap[1] = min(r1[1], r2[1])
            return self._omega_function(r1, overlap)

    def _omega_function(self, rrange, overlap):
        anomaly_length = rrange[1] - rrange[0] + 1
        my_positional_bias = 0
        max_positional_bias = 0
        temp_bias = 0
        for i in range(1, anomaly_length + 1):
            temp_bias = self._delta_function(i, anomaly_length)
            max_positional_bias += temp_bias
            j = rrange[0] + i - 1
            if j >= overlap[0] and j <= overlap[1]:
                my_positional_bias += temp_bias
        if max_positional_bias > 0:
            res = my_positional_bias / max_positional_bias
            return res
        else:
            return 0

    def _delta_function(self, t, anomaly_length):
        return self._delta_select(self.bias, t, anomaly_length)

    def _delta_select(self, delta, t, anomaly_length):
        if delta == "flat":
            return 1.0
        elif delta == "front":
            return float(anomaly_length - t + 1.0)
        elif delta == "middle":
            if t <= anomaly_length / 2.0:
                return float(t)
            else:
                return float(anomaly_length - t + 1.0)
        elif delta == "back":
            return float(t)
        elif delta == "udf_delta":
            return self._udf_delta(t, anomaly_length)
        else:
            raise Exception("Invalid positional bias value")

    def _udf_delta(self):
        """
        user defined delta function
        """

        return 1.0

    def _shift(self, arr, num, fill_value=np.nan):
        arr = np.roll(arr, num)
        if num < 0:
            arr[num:] = fill_value
        elif num > 0:
            arr[:num] = fill_value
        return arr

    def _prepare_data(self, values_real, values_pred):

        assert len(values_real) == len(values_pred)
        assert np.allclose(np.unique(values_real), np.array([0, 1])) or np.allclose(
            np.unique(values_real), np.array([1])
        )
        assert np.allclose(np.unique(values_pred), np.array([0, 1])) or np.allclose(
            np.unique(values_pred), np.array([1])
        )

        predicted_anomalies_ = np.argwhere(values_pred == 1).ravel()
        predicted_anomalies_shift_forward = self._shift(predicted_anomalies_, 1, fill_value=predicted_anomalies_[0])
        predicted_anomalies_shift_backward = self._shift(predicted_anomalies_, -1, fill_value=predicted_anomalies_[-1])
        predicted_anomalies_start = np.argwhere(
            (predicted_anomalies_shift_forward - predicted_anomalies_) != -1
        ).ravel()
        predicted_anomalies_finish = np.argwhere(
            (predicted_anomalies_ - predicted_anomalies_shift_backward) != -1
        ).ravel()
        predicted_anomalies = np.hstack(
            [
                predicted_anomalies_[predicted_anomalies_start].reshape(-1, 1),
                predicted_anomalies_[predicted_anomalies_finish].reshape(-1, 1),
            ]
        )

        real_anomalies_ = np.argwhere(values_real == 1).ravel()
        real_anomalies_shift_forward = self._shift(real_anomalies_, 1, fill_value=real_anomalies_[0])
        real_anomalies_shift_backward = self._shift(real_anomalies_, -1, fill_value=real_anomalies_[-1])
        real_anomalies_start = np.argwhere((real_anomalies_shift_forward - real_anomalies_) != -1).ravel()
        real_anomalies_finish = np.argwhere((real_anomalies_ - real_anomalies_shift_backward) != -1).ravel()
        real_anomalies = np.hstack(
            [
                real_anomalies_[real_anomalies_start].reshape(-1, 1),
                real_anomalies_[real_anomalies_finish].reshape(-1, 1),
            ]
        )

        return real_anomalies, predicted_anomalies



from .time_series_metrics.fscore import TimeSeriesFScore
from .time_series_metrics.precision import TimeSeriesPrecision
from .time_series_metrics.recall import TimeSeriesRecall


def ts_precision(real, pred, alpha=0.0, cardinality="one", bias="flat"):
    """Compute the range based precision.

    The range based precision is the average of "Precision_Ti", where "Precision_Ti" is
    the precision score of each predicted anomaly range.
    "Precision_Ti" for a single predicted anomaly range is calculated by the following formula.
        Precision_Ti = α x ExistenceReward + (1 - α) x OverlapReward , where 0 ≤ α ≤ 1
    α represents the relative importance of rewarding existence, whereas
    (1 − α) represents the relative importance of rewarding size, position, and cardinality.

    "ExistenceReward" is 1 if a real anomaly range has overlap with even a single point of
    the predicted anomaly range, 0 otherwise.
    Note: For prediction, there is no need for an existence reward, since precision by definition
          emphasizes prediction quality, and existence by itself is too low a bar for judging
          the quality of a prediction (i.e., α = 0).

    "OverlapReward" is calculated by the following formula.
        OverlapReward = CardinalityFactor x Sum of ω
    "CardinalityFactor" is 1 if the predicted anomaly range overlaps with only one real anomaly range.
    Otherwise it receives 0 ≤ γ() ≤ 1 defined by the application.
    "CardinalityFactor" serves as a scaling factor for the rewards "ω"s, which is earned from overlap
    size and position.
    In determing "ω", we consider the size of the correctly predicted portion of an predicted anomaly
    range and the relative position of the correctly predicted portion of an predicted anomaly range.

    Args:
        real: np.ndarray
            One-dimensional array of correct answers with values of 1 or 0.
        pred: np.ndarray
            One-dimensional array of predicted answers with values of 1 or 0.
        alpha: float, default=0.0
            Relative importance of existence reward. 0 ≤ alpha ≤ 1.
        cardinality: string, default="one"
            Cardinality type. This should be "one", "reciprocal" or "udf_gamma".
        bias: string, default="flat"
            Positional bias. This should be "flat", "front", "middle", or "back"

    Returns:
        float: precision.score
    """
    precision = TimeSeriesPrecision(alpha, cardinality, bias)
    return precision.score(real, pred)


def ts_recall(real, pred, alpha=0.0, cardinality="one", bias="flat"):
    """Compute the range based recall.

    The range based recall is the average of "Recall_Ti", where "Recall_Ti" is
    the recall score of each real anomaly range.
    "Recall_Ti" for a single real anomaly range is calculated by the following formula.
        Recall_Ti = α x ExistenceReward + (1 - α) x OverlapReward , where 0 ≤ α ≤ 1
    α represents the relative importance of rewarding existence, whereas
    (1 − α) represents the relative importance of rewarding size, position, and cardinality.

    "ExistenceReward" is 1 if a prediction captures even a single point of the real anomaly range, 0 otherwise.

    "OverlapReward" is calculated by the following formula.
        OverlapReward = CardinalityFactor x Sum of ω
    "CardinalityFactor" is 1 if the real anomaly range overlaps with only one predicted anomaly range.
    Otherwise it receives 0 ≤ γ() ≤ 1 defined by the application.
    "CardinalityFactor" serves as a scaling factor for the rewards "ω"s, which is earned from overlap
    size and position.
    In determing "ω", we consider the size of the correctly predicted portion of the real anomaly range
    and the relative
    position of the correctly predicted portion of the real anomaly range.

    Args:
        real: np.ndarray
            One-dimensional array of correct answers with values of 1 or 0.
        pred: np.ndarray
            One-dimensional array of predicted answers with values of 1 or 0.
        alpha: float, default=0.0
            Relative importance of existence reward. 0 ≤ alpha ≤ 1.
        cardinality: string, default="one"
            Cardinality type. This should be "one", "reciprocal" or "udf_gamma".
        bias: string, default="flat"
            Positional bias. This should be "flat", "front", "middle", or "back"

    Returns:
        float: recall.score
    """
    recall = TimeSeriesRecall(alpha, cardinality, bias)
    return recall.score(real, pred)


def ts_fscore(real, pred, beta=1.0, p_alpha=0.0, r_alpha=0.0, cardinality="one", p_bias="flat", r_bias="flat"):
    """Compute the range based f-score

    The F-beta score is the weighted harmonic mean of precision and recall,
    reaching its optimal value at 1 and its worst value at 0.
    The beta parameter determines the weight of recall in the combined score.
    beta < 1 lends more weight to precision, while beta > 1 favors recall
    (beta -> 0 considers only precision, beta -> +inf only recall).

    Args:
        real: np.ndarray
            One-dimensional array of correct answers with values of 1 or 0.
        pred: np.ndarray
            One-dimensional array of predicted answers with values of 1 or 0.
        p_alpha: float, default=0.0
            Relative importance of existence reward for precision. 0 ≤ alpha ≤ 1.
        r_alpha: float, default=0.0
            Relative importance of existence reward for recall. 0 ≤ alpha ≤ 1.
        cardinality: string, default="one"
            Cardinality type. This should be "one", "reciprocal" or "udf_gamma".
        p_bias: string, default="flat"
            Positional bias for precision. This should be "flat", "front", "middle", or "back"
        r_bias: string, default="flat"
            Positional bias for recall. This should be "flat", "front", "middle", or "back"

    Returns:
        float: f.score
    """

    fscore = TimeSeriesFScore(beta, p_alpha, r_alpha, cardinality, p_bias, r_bias)
    return fscore.score(real, pred)
